/*
 * Copyright (C) 2016 AriaLyy(https://github.com/AriaLyy/Aria)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.arialyy.aria.core.scheduler;

import android.content.Intent;
import android.os.Bundle;
import android.os.CountDownTimer;
import android.os.Message;
import com.arialyy.aria.core.AriaManager;
import com.arialyy.aria.core.download.DownloadGroupTask;
import com.arialyy.aria.core.download.DownloadTask;
import com.arialyy.aria.core.inf.AbsEntity;
import com.arialyy.aria.core.inf.AbsNormalEntity;
import com.arialyy.aria.core.inf.AbsTask;
import com.arialyy.aria.core.inf.AbsTaskWrapper;
import com.arialyy.aria.core.inf.GroupSendParams;
import com.arialyy.aria.core.inf.IEntity;
import com.arialyy.aria.core.inf.ITask;
import com.arialyy.aria.core.inf.TaskSchedulerType;
import com.arialyy.aria.core.manager.TaskWrapperManager;
import com.arialyy.aria.core.queue.ITaskQueue;
import com.arialyy.aria.core.upload.UploadTask;
import com.arialyy.aria.util.ALog;
import com.arialyy.aria.util.NetUtils;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ScheduledThreadPoolExecutor;
import java.util.concurrent.TimeUnit;

/**
 * Created by lyy on 2017/6/4. 事件调度器，用于处理任务状态的调度
 */
abstract class AbsSchedulers<TASK_ENTITY extends AbsTaskWrapper, TASK extends ITask,
    QUEUE extends ITaskQueue<TASK, TASK_ENTITY>> implements ISchedulers {
  private final String TAG = "AbsSchedulers";

  protected QUEUE mQueue;

  private Map<String, AbsSchedulerListener<TASK, AbsNormalEntity>> mObservers =
      new ConcurrentHashMap<>();

  /**
   * 设置代理类后缀名
   */
  abstract String getProxySuffix();

  private AriaManager manager;

  AbsSchedulers() {
    manager = AriaManager.getInstance(AriaManager.APP);
  }

  /**
   * 将当前类注册到Aria
   *
   * @param obj 观察者类
   */
  public void register(Object obj) {
    String targetName = obj.getClass().getName();
    AbsSchedulerListener<TASK, AbsNormalEntity> listener = mObservers.get(getKey(obj));
    if (listener == null) {
      listener = createListener(targetName);
      if (listener != null) {
        listener.setListener(obj);
        mObservers.put(getKey(obj), listener);
      } else {
        ALog.e(TAG, "注册错误，没有【" + targetName + "】观察者");
      }
    }
  }

  /**
   * 移除注册
   *
   * @param obj 观察者类
   */
  public void unRegister(Object obj) {
    if (!mObservers.containsKey(getKey(obj))) {
      return;
    }
    for (Iterator<Map.Entry<String, AbsSchedulerListener<TASK, AbsNormalEntity>>> iter =
        mObservers.entrySet().iterator(); iter.hasNext(); ) {
      Map.Entry<String, AbsSchedulerListener<TASK, AbsNormalEntity>> entry = iter.next();
      if (entry.getKey().equals(getKey(obj))) {
        iter.remove();
      }
    }
  }

  private String getKey(Object obj) {
    return obj.getClass().getName() + obj.hashCode();
  }

  /**
   * 创建代理类
   *
   * @param targetName 通过观察者创建对应的Aria事件代理
   */
  private AbsSchedulerListener<TASK, AbsNormalEntity> createListener(String targetName) {
    AbsSchedulerListener<TASK, AbsNormalEntity> listener = null;
    try {
      Class clazz = Class.forName(targetName + getProxySuffix());
      listener = (AbsSchedulerListener<TASK, AbsNormalEntity>) clazz.newInstance();
    } catch (ClassNotFoundException e) {
      ALog.e(TAG, targetName + "，没有Aria的Download或Upload注解方法");
    } catch (InstantiationException e) {
      ALog.e(TAG, e.getMessage());
    } catch (IllegalAccessException e) {
      ALog.e(TAG, e.getMessage());
    }
    return listener;
  }

  @Override public boolean handleMessage(Message msg) {
    if (msg.arg1 == IS_SUB_TASK) {
      return handleSubEvent(msg);
    }

    TASK task = (TASK) msg.obj;
    if (task == null) {
      ALog.e(TAG, "请传入下载任务");
      return true;
    }
    handleNormalEvent(task, msg.what);
    return true;
  }

  /**
   * 处理任务组子任务事件
   */
  private boolean handleSubEvent(Message msg) {
    GroupSendParams params = (GroupSendParams) msg.obj;
    if (mObservers.size() > 0) {
      Set<String> keys = mObservers.keySet();
      for (String key : keys) {
        AbsSchedulerListener<TASK, AbsNormalEntity> listener = mObservers.get(key);
        switch (msg.what) {
          case SUB_PRE:
            listener.onSubTaskPre((TASK) params.groupTask, params.entity);
            break;
          case SUB_START:
            listener.onSubTaskStart((TASK) params.groupTask, params.entity);
            break;
          case SUB_STOP:
            listener.onSubTaskStop((TASK) params.groupTask, params.entity);
            break;
          case SUB_FAIL:
            listener.onSubTaskFail((TASK) params.groupTask, params.entity,
                (Exception) (params.groupTask).getExpand(AbsTask.ERROR_INFO_KEY));
            break;
          case SUB_RUNNING:
            listener.onSubTaskRunning((TASK) params.groupTask, params.entity);
            break;
          case SUB_CANCEL:
            listener.onSubTaskCancel((TASK) params.groupTask, params.entity);
            break;
          case SUB_COMPLETE:
            listener.onSubTaskComplete((TASK) params.groupTask, params.entity);
            break;
        }
      }
    }

    boolean canSend = manager.getAppConfig().isUseBroadcast();
    if (canSend) {
      AriaManager.APP.sendBroadcast(
          createData(msg.what, ITask.DOWNLOAD_GROUP_SUB, params.entity));
    }

    return true;
  }

  /**
   * 处理普通任务和任务组的事件
   */
  private void handleNormalEvent(TASK task, int what) {
    switch (what) {
      case STOP:
        if (task.getState() == IEntity.STATE_WAIT) {
          break;
        }
        mQueue.removeTaskFormQueue(task.getKey());
        if (mQueue.getCurrentExePoolNum() < mQueue.getMaxTaskNum()) {
          ALog.d(TAG, String.format("停止任务【%s】成功，尝试开始下一任务", task.getTaskName()));
          startNextTask(task);
        } else {
          ALog.d(TAG, String.format("停止任务【%s】成功", task.getTaskName()));
        }
        break;
      case CANCEL:
        mQueue.removeTaskFormQueue(task.getKey());
        if (mQueue.getCurrentExePoolNum() < mQueue.getMaxTaskNum()) {
          ALog.d(TAG, String.format("删除任务【%s】成功，尝试开始下一任务", task.getTaskName()));
          startNextTask(task);
        } else {
          ALog.d(TAG, String.format("删除任务【%s】成功", task.getTaskName()));
        }
        break;
      case COMPLETE:
        mQueue.removeTaskFormQueue(task.getKey());
        ALog.d(TAG, String.format("任务【%s】处理完成", task.getTaskName()));
        startNextTask(task);
        break;
      case FAIL:
        handleFailTask(task);
        break;
    }
    if (what == CANCEL || what == COMPLETE) {
      TaskWrapperManager.getInstance().removeTaskWrapper(task.getKey());
    } else {
      if (what != RUNNING) {
        TaskWrapperManager.getInstance().putTaskWrapper(task.getKey(), task.getTaskWrapper());
      }
    }
    if (what != FAIL) {
      callback(what, task);
    }
  }

  /**
   * 回调
   *
   * @param state 状态
   */
  private void callback(int state, TASK task) {
    sendNormalBroadcast(state, task);
    if (mObservers.size() > 0) {
      Set<String> keys = mObservers.keySet();
      for (String key : keys) {
        callback(state, task, mObservers.get(key));
      }
    }
  }

  private void callback(int state, TASK task,
      AbsSchedulerListener<TASK, AbsNormalEntity> listener) {
    if (listener != null) {
      if (task == null) {
        ALog.e(TAG, "TASK 为null，回调失败");
        return;
      }
      switch (state) {
        case WAIT:
          listener.onWait(task);
          break;
        case PRE:
          listener.onPre(task);
          break;
        case POST_PRE:
          listener.onTaskPre(task);
          break;
        case RUNNING:
          listener.onTaskRunning(task);
          break;
        case START:
          listener.onTaskStart(task);
          break;
        case STOP:
          listener.onTaskStop(task);
          break;
        case RESUME:
          listener.onTaskResume(task);
          break;
        case CANCEL:
          listener.onTaskCancel(task);
          break;
        case COMPLETE:
          listener.onTaskComplete(task);
          break;
        case FAIL:
          listener.onTaskFail(task.getTaskType() == ITask.TEMP ? null : task,
              (Exception) task.getExpand(AbsTask.ERROR_INFO_KEY));
          break;
        case NO_SUPPORT_BREAK_POINT:
          listener.onNoSupportBreakPoint(task);
          break;
      }
    }
  }

  /**
   * 发送普通任务的广播
   */
  private void sendNormalBroadcast(int state, TASK task) {
    boolean canSend = manager.getAppConfig().isUseBroadcast();
    if (!canSend) {
      return;
    }
    int type = task.getTaskType();
    if (type == ITask.DOWNLOAD || type == ITask.DOWNLOAD_GROUP) {
      AriaManager.APP.sendBroadcast(
          createData(state, type, task.getTaskWrapper().getEntity()));
    } else if (type == ITask.UPLOAD) {
      AriaManager.APP.sendBroadcast(
          createData(state, type, task.getTaskWrapper().getEntity()));
    } else {
      ALog.w(TAG, "发送广播失败，没有对应的任务");
    }
  }

  /**
   * 创建广播发送的数据
   *
   * @param taskState 任务状态 {@link ISchedulers}
   * @param taskType 任务类型 {@link ITask}
   * @param entity 任务实体
   */
  private Intent createData(int taskState, int taskType, AbsEntity entity) {
    Intent intent = new Intent(ISchedulers.ARIA_TASK_INFO_ACTION);
    Bundle b = new Bundle();
    b.putInt(ISchedulers.TASK_TYPE, taskType);
    b.putInt(ISchedulers.TASK_STATE, taskState);
    b.putLong(ISchedulers.TASK_SPEED, entity.getSpeed());
    b.putInt(ISchedulers.TASK_PERCENT, entity.getPercent());
    b.putParcelable(ISchedulers.TASK_ENTITY, entity);
    intent.putExtras(b);
    return intent;
  }

  /**
   * 处理下载任务下载失败的情形
   *
   * @param task 下载任务
   */
  private void handleFailTask(final TASK task) {
    if (!task.isNeedRetry() || task.isStop() || task.isCancel()) {
      mQueue.removeTaskFormQueue(task.getKey());
      startNextTask(task);
      callback(FAIL, task);
      return;
    }
    long interval = task.getTaskWrapper().getConfig().getReTryInterval();
    int num = task.getTaskWrapper().getConfig().getReTryNum();
    boolean isNotNetRetry = manager.getAppConfig().isNotNetRetry();

    final int reTryNum = num;
    if ((!NetUtils.isConnected(AriaManager.APP) && !isNotNetRetry)
        || task.getTaskWrapper().getEntity().getFailNum() > reTryNum) {
      mQueue.removeTaskFormQueue(task.getKey());
      startNextTask(task);
      TaskWrapperManager.getInstance().removeTaskWrapper(task.getKey());
      callback(FAIL, task);
      return;
    }

    final ScheduledThreadPoolExecutor timer = new ScheduledThreadPoolExecutor(1);

    timer.schedule(new Runnable() {
      @Override public void run() {
        AbsEntity entity = task.getTaskWrapper().getEntity();
        if (entity.getFailNum() <= reTryNum) {
          ALog.d(TAG, String.format("任务【%s】开始重试", task.getTaskName()));
          TASK task = mQueue.getTask(entity.getKey());
          mQueue.reTryStart(task);
        } else {
          mQueue.removeTaskFormQueue(task.getKey());
          startNextTask(task);
          TaskWrapperManager.getInstance().removeTaskWrapper(task.getKey());
        }
      }
    }, interval, TimeUnit.MILLISECONDS);
  }

  /**
   * 启动下一个任务，条件：任务停止，取消下载，任务完成
   */
  private void startNextTask(TASK oldTask) {
    if (oldTask.getSchedulerType() == TaskSchedulerType.TYPE_STOP_NOT_NEXT) {
      return;
    }
    TASK newTask = mQueue.getNextTask();
    if (newTask == null) {
      if (mQueue.getCurrentExePoolNum() == 0) {
        ALog.i(TAG, "没有等待中的任务");
      }
      return;
    }
    if (newTask.getState() == IEntity.STATE_WAIT) {
      mQueue.startTask(newTask);
    }
  }
}
